""" BitBlock
    Copyright (C) 2022  C. Lockett

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

class ConfigurationError(Exception):
    """ Exception raised when the RPCConnection has been configured 
    improperly.

    """
    pass


class RPCError(Exception):
    """ Exception raised when the Bitcoin RPC returns an error. """
    def __init__(self, code: int, message: str) -> None:
        self.code = code
        self.message = message